

# Porting of resnet2D to 3D conv and is also able to keep time dimension through blocks/bottlenecks


import tflearn
import tensorflow as tf
from tflearn.layers import conv_3d
from tflearn import utils

def residual_block3D(incoming, nb_blocks, out_channels, downsample=False, keep_time_dim=False, print_Layer=False,
                     type='B', downsample_strides=2, activation='relu', batch_norm=True,
                     bias=True, weights_init='variance_scaling',
                     bias_init='zeros', regularizer='L2', weight_decay=0.0001,
                     trainable=True, restore=True, reuse=False, scope=None,
                     name="ResidualBlock3D"):

    """
    A 3D-residual block extended from the residual block described in MSRA's Deep Residual Network paper. Full pre-activation architecture is used here. (This implementation is an extension of tflearn.layers.conv.residual_block)

    :param incoming: 5-D Tensor [batch, numFrames, height, width, in_channels].

    :param nb_blocks: 5-D Tensor [batch, numFrames, new height, new width, nb_filter].

    :param bottleneck_size: int`. The number of convolutional filter of the bottleneck convolutional layer.

    :param out_channels: `int`. The number of convolutional filters of the layers surrounding the bottleneck layer.

    :param keep_time_dim: 'boolean'. Tells us if we will keep temporal dimension in this layer.

    :param print_Layer: 'boolean'. Tells us if we will printin this layer.

    :param type: 'char'. Tells us which downsampling strategy we will use.

    :param downsample: `bool`. If True, apply downsampling using 'downsample_strides' for strides.

    :param downsample_strides: int`. The strides to use when downsampling.

    :param activation: `str` (name) or `function` (returning a `Tensor`). Activation applied to this layer (see tflearn.activations). Default: 'linear'.

    :param batch_norm: `bool`. If True, apply batch normalization.

    :param bias: `bool`. If True, a bias is used.

    :param weights_init: `str` (name) or `Tensor`. Weights initialization. (see tflearn.initializations) Default: 'uniform_scaling'.

    :param bias_init: `str` (name) or `tf.Tensor`. Bias initialization. (see tflearn.initializations) Default: 'zeros'.

    :param regularizer: `str` (name) or `Tensor`. Add a regularizer to this layer weights (see tflearn.regularizers). Default: None.

    :param weight_decay: `float`. Regularizer decay parameter. Default: 0.001.

    :param trainable: `bool`. If True, weights will be trainable.

    :param restore: `bool`. If True, this layer weights will be restored when loading a model.

    :param reuse: `bool`. If True and 'scope' is provided, this layer variables will be reused (shared).

    :param scope: `str`. Define this layer scope (optional). A scope can be used to share variables between layers. Note that scope will override name.

    :param name: A name for this layer (optional). Default: 'DeepBottleneck'.

    :return: 5-D Tensor [batch, numFrames, new height, new width, nb_filter].
    """
    resnet = incoming
    in_channels = incoming.get_shape().as_list()[-1]

    # Variable Scope fix for older TF
    with tf.variable_scope(scope, default_name=name, values=[incoming],
                           reuse=reuse) as scope:

        name = scope.name #TODO

        for i in range(nb_blocks):
            identity = resnet

            if print_Layer:
                print resnet.shape

            if not downsample:
                downsample_strides = 1

            if batch_norm:
                resnet = tflearn.batch_normalization(resnet)

            resnet = tflearn.activation(resnet, activation)

            # CONV1 3x3

            if keep_time_dim:
                resnet = conv_3d(resnet, out_channels, 3,
                                 [1, downsample_strides, downsample_strides], 'same', 'linear',
                                 bias, weights_init, bias_init,
                                 regularizer, weight_decay, trainable,
                                 restore)
            else:
                resnet = conv_3d(resnet, out_channels, 3,
                                 downsample_strides, 'same', 'linear',
                                 bias, weights_init, bias_init,
                                 regularizer, weight_decay, trainable,
                                 restore)

            if batch_norm:
                resnet = tflearn.batch_normalization(resnet)

            resnet = tflearn.activation(resnet, activation)

            # CONV2 3x3

            resnet = conv_3d(resnet, out_channels, 3,
                             1, 'same', 'linear',
                             bias, weights_init, bias_init,
                             regularizer, weight_decay, trainable,
                             restore)
            # Downsampling

            if downsample_strides > 1:
                # Conv with filter size 1
                if type == 'C':
                    if keep_time_dim:
                        identity = tflearn.conv_3d(identity, out_channels, 1,
                                                   [1, downsample_strides, downsample_strides], 'same', 'linear',
                                                   bias, weights_init, bias_init,
                                                   regularizer, weight_decay, trainable,
                                                   restore)

                    else:
                        identity = tflearn.conv_3d(identity, out_channels, 1,
                                                   downsample_strides, 'same', 'linear',
                                                   bias, weights_init, bias_init,
                                                   regularizer, weight_decay, trainable,
                                                   restore)
                elif type == 'B':

                    if keep_time_dim:
                        identity = tflearn.avg_pool_3d(identity, [1, 1, 2, 2, 1], strides=[1, 2, 2])
                    else:
                        identity = tflearn.avg_pool_3d(identity, downsample_strides,
                                                  downsample_strides)

                elif type == 'A':

                    if keep_time_dim:
                        identity = tflearn.max_pool_3d(identity, [1, 1, 2, 2, 1], strides=[1, 2, 2])
                    else:
                        identity = tflearn.max_pool_3d(identity, downsample_strides,
                                                       downsample_strides)


            # Projection to new dimension
            if type == 'B' or type =='A':
                if in_channels != out_channels:
                    ch = (out_channels - in_channels)//2
                    identity = tf.pad(identity,
                                      [[0, 0], [0, 0], [0, 0], [0, 0], [ch, ch]])
                    in_channels = out_channels

            resnet = resnet + identity

    return resnet


def residual_bottleneck3D(incoming, nb_blocks, bottleneck_size, out_channels, keep_time_dim = False, print_Layer=False,
                          type='B', downsample=False, downsample_strides=2,
                          activation='relu', batch_norm=True, bias=True,
                          weights_init='variance_scaling', bias_init='zeros',
                          regularizer='L2', weight_decay=0.0001,
                          trainable=True, restore=True, reuse=False, scope=None,
                          name="ResidualBottleneck3D"):

    """
    A 3D-residual bottleneck block extended from the residual bottleneck described in MSRA's Deep Residual Network paper. Full pre-activation architecture is used here. (This implementation is an extension of tflearn.layers.conv.residual_bottleneck)

    :param incoming: 5-D Tensor [batch, numFrames, height, width, in_channels].

    :param nb_blocks: 5-D Tensor [batch, numFrames, new height, new width, nb_filter].

    :param bottleneck_size: int`. The number of convolutional filter of the bottleneck convolutional layer.

    :param out_channels: `int`. The number of convolutional filters of the layers surrounding the bottleneck layer.

    :param keep_time_dim: 'boolean'. Tells us if we will keep temporal dimension in this layer.

    :param print_Layer: 'boolean'. Tells us if we will printin this layer.

    :param type: 'char'. Tells us which downsampling strategy we will use.

    :param downsample: `bool`. If True, apply downsampling using 'downsample_strides' for strides.

    :param downsample_strides: int`. The strides to use when downsampling.

    :param activation: `str` (name) or `function` (returning a `Tensor`). Activation applied to this layer (see tflearn.activations). Default: 'linear'.

    :param batch_norm: `bool`. If True, apply batch normalization.

    :param bias: `bool`. If True, a bias is used.

    :param weights_init: `str` (name) or `Tensor`. Weights initialization. (see tflearn.initializations) Default: 'uniform_scaling'.

    :param bias_init: `str` (name) or `tf.Tensor`. Bias initialization. (see tflearn.initializations) Default: 'zeros'.

    :param regularizer: `str` (name) or `Tensor`. Add a regularizer to this layer weights (see tflearn.regularizers). Default: None.

    :param weight_decay: `float`. Regularizer decay parameter. Default: 0.001.

    :param trainable: `bool`. If True, weights will be trainable.

    :param restore: `bool`. If True, this layer weights will be restored when loading a model.

    :param reuse: `bool`. If True and 'scope' is provided, this layer variables will be reused (shared).

    :param scope: `str`. Define this layer scope (optional). A scope can be used to share variables between layers. Note that scope will override name.

    :param name: A name for this layer (optional). Default: 'DeepBottleneck'.

    :return: 5-D Tensor [batch, numFrames, new height, new width, nb_filter].
    """
    resnet = incoming
    in_channels = incoming.get_shape().as_list()[-1]

    with tf.variable_scope(scope, default_name=name, values=[incoming],
                           reuse=reuse) as scope:

        name = scope.name #TODO

        for i in range(nb_blocks):

            identity = resnet
            print "Identity " + str(identity.shape)

            if print_Layer:
                print resnet.shape

            if not downsample:
                downsample_strides = 1

            if batch_norm:
                resnet = tflearn.batch_normalization(resnet)
            resnet = tflearn.activation(resnet, activation)
            print "Input: " + str(resnet.shape)
            if keep_time_dim:
                resnet = conv_3d(resnet, bottleneck_size, 1,
                                 [1, downsample_strides, downsample_strides], 'valid',
                                 'linear', bias, weights_init,
                                 bias_init, regularizer, weight_decay,
                                 trainable, restore)
            else:
                resnet = conv_3d(resnet, bottleneck_size, 1,
                                 downsample_strides, 'valid',
                                 'linear', bias, weights_init,
                                 bias_init, regularizer, weight_decay,
                                 trainable, restore)
            print resnet.shape
            print "Downsample" + str(resnet.shape)

            if batch_norm:
                resnet = tflearn.batch_normalization(resnet)
            resnet = tflearn.activation(resnet, activation)

            resnet = conv_3d(resnet, bottleneck_size, 3, 1, 'same',
                             'linear', bias, weights_init,
                             bias_init, regularizer, weight_decay,
                             trainable, restore)

            print "3x3 conv " + str(resnet.shape)

            resnet = conv_3d(resnet, out_channels, 1, 1, 'valid',
                             activation, bias, weights_init,
                             bias_init, regularizer, weight_decay,
                             trainable, restore)

            print "Upsample " + str(resnet.shape)

            # Downsampling
            if downsample_strides > 1:
                # Conv with filter size 1
                if type == 'C':
                    if keep_time_dim:
                        identity = tflearn.conv_3d(identity, out_channels, 1,
                                                   [1, downsample_strides, downsample_strides], 'same', 'linear',
                                                   bias, weights_init, bias_init,
                                                   regularizer, weight_decay, trainable,
                                                   restore)

                    else:
                        identity = tflearn.conv_3d(identity, out_channels, 1,
                                                   downsample_strides, 'same', 'linear',
                                                   bias, weights_init, bias_init,
                                                   regularizer, weight_decay, trainable,
                                                   restore)
                elif type == 'B':

                    if keep_time_dim:
                        identity = tflearn.avg_pool_3d(identity, [1, 1, 2, 2, 1], strides=[1, 2, 2])
                    else:
                        identity = tflearn.avg_pool_3d(identity, downsample_strides,
                                                       downsample_strides)

            # Projection to new dimension
            if type == 'B':
                if in_channels != out_channels:
                    ch = (out_channels - in_channels) // 2
                    identity = tf.pad(identity,
                                      [[0, 0], [0, 0], [0, 0], [0, 0], [ch, ch]])
                    in_channels = out_channels

                resnet = resnet + identity

    return resnet

def LongResidual_block3D(incoming, nb_blocks, out_channels, downsample=False, keep_time_dim=False, print_Layer=False,
                     type='B', downsample_strides=2, activation='relu', batch_norm=True,
                     bias=True, weights_init='variance_scaling',
                     bias_init='zeros', regularizer='L2', weight_decay=0.0001,
                     trainable=True, restore=True, reuse=False, scope=None,
                     name="LongResidualBlock3D"):
    """
    A long 3D-residual block extended from the residual block described in MSRA's Deep Residual Network paper. Full pre-activation architecture is used here. (This implementation is an extension of tflearn.layers.conv.residual_block)

    :param incoming: 5-D Tensor [batch, numFrames, height, width, in_channels].

    :param nb_blocks: 5-D Tensor [batch, numFrames, new height, new width, nb_filter].

    :param bottleneck_size: int`. The number of convolutional filter of the bottleneck convolutional layer.

    :param out_channels: `int`. The number of convolutional filters of the layers surrounding the bottleneck layer.

    :param keep_time_dim: 'boolean'. Tells us if we will keep temporal dimension in this layer.

    :param print_Layer: 'boolean'. Tells us if we will printin this layer.

    :param type: 'char'. Tells us which downsampling strategy we will use.

    :param downsample: `bool`. If True, apply downsampling using 'downsample_strides' for strides.

    :param downsample_strides: int`. The strides to use when downsampling.

    :param activation: `str` (name) or `function` (returning a `Tensor`). Activation applied to this layer (see tflearn.activations). Default: 'linear'.

    :param batch_norm: `bool`. If True, apply batch normalization.

    :param bias: `bool`. If True, a bias is used.

    :param weights_init: `str` (name) or `Tensor`. Weights initialization. (see tflearn.initializations) Default: 'uniform_scaling'.

    :param bias_init: `str` (name) or `tf.Tensor`. Bias initialization. (see tflearn.initializations) Default: 'zeros'.

    :param regularizer: `str` (name) or `Tensor`. Add a regularizer to this layer weights (see tflearn.regularizers). Default: None.

    :param weight_decay: `float`. Regularizer decay parameter. Default: 0.001.

    :param trainable: `bool`. If True, weights will be trainable.

    :param restore: `bool`. If True, this layer weights will be restored when loading a model.

    :param reuse: `bool`. If True and 'scope' is provided, this layer variables will be reused (shared).

    :param scope: `str`. Define this layer scope (optional). A scope can be used to share variables between layers. Note that scope will override name.

    :param name: A name for this layer (optional). Default: 'DeepBottleneck'.

    :return: 5-D Tensor [batch, numFrames, new height, new width, nb_filter].
    """
    resnet = incoming
    in_channels = incoming.get_shape().as_list()[-1]

    # Variable Scope fix for older TF
    with tf.variable_scope(scope, default_name=name, values=[incoming],
                           reuse=reuse) as scope:

        name = scope.name #TODO
        LongResidual = 0
        if not downsample:
            LongResidual = resnet

        for i in range(nb_blocks):
            identity = resnet

            if print_Layer:
                print resnet.shape

            if not downsample:
                downsample_strides = 1

            if batch_norm:
                resnet = tflearn.batch_normalization(resnet)

            resnet = tflearn.activation(resnet, activation)

            if keep_time_dim:
                resnet = conv_3d(resnet, out_channels, 3,
                                 [1, downsample_strides, downsample_strides], 'same', 'linear',
                                 bias, weights_init, bias_init,
                                 regularizer, weight_decay, trainable,
                                 restore)
            else:
                resnet = conv_3d(resnet, out_channels, 3,
                                 downsample_strides, 'same', 'linear',
                                 bias, weights_init, bias_init,
                                 regularizer, weight_decay, trainable,
                                 restore)

            if batch_norm:
                resnet = tflearn.batch_normalization(resnet)

            resnet = tflearn.activation(resnet, activation)


            resnet = conv_3d(resnet, out_channels, 3,
                             1, 'same', 'linear',
                             bias, weights_init, bias_init,
                             regularizer, weight_decay, trainable,
                             restore)
            # Downsampling

            if downsample_strides > 1:
                # Conv with filter size 1
                if type == 'C':
                    if keep_time_dim:
                        identity = tflearn.conv_3d(identity, out_channels, 1,
                                                   [1, downsample_strides, downsample_strides], 'same', 'linear',
                                                   bias, weights_init, bias_init,
                                                   regularizer, weight_decay, trainable,
                                                   restore)

                    else:
                        identity = tflearn.conv_3d(identity, out_channels, 1,
                                                   downsample_strides, 'same', 'linear',
                                                   bias, weights_init, bias_init,
                                                   regularizer, weight_decay, trainable,
                                                   restore)
                elif type == 'B':

                    if keep_time_dim:
                        identity = tflearn.avg_pool_3d(identity, [1, 1, 2, 2, 1], strides=[1, 2, 2])
                    else:
                        identity = tflearn.avg_pool_3d(identity, downsample_strides,
                                                  downsample_strides)

                elif type == 'A':

                    if keep_time_dim:
                            identity = tflearn.max_pool_3d(identity, [1, 1, 2, 2, 1], strides=[1, 2, 2])
                    else:
                        identity = tflearn.max_pool_3d(identity, downsample_strides,
                                                           downsample_strides)




            # Projection to new dimension
            if type == 'B' or type == 'A':
                if in_channels != out_channels:
                    ch = (out_channels - in_channels)//2
                    identity = tf.pad(identity,
                                      [[0, 0], [0, 0], [0, 0], [0, 0], [ch, ch]])
                    in_channels = out_channels

            resnet = resnet + identity


        resnet = resnet + LongResidual

    return resnet


def long_residual_bottleneck3D(incoming, nb_blocks, bottleneck_size, out_channels, keep_time_dim = False, print_Layer=False,
                          type='B', downsample=False, downsample_strides=2,
                          activation='relu', batch_norm=True, bias=True,
                          weights_init='variance_scaling', bias_init='zeros',
                          regularizer='L2', weight_decay=0.0001,
                          trainable=True, restore=True, reuse=False, scope=None,
                          name="LongResidualBottleneck3D"):
    """
    A 3D-residual bottleneck block extended from the residual bottleneck described in MSRA's Deep Residual Network paper. Full pre-activation architecture is used here. (This implementation is an extension of tflearn.layers.conv.residual_bottleneck)

    :param incoming: 5-D Tensor [batch, numFrames, height, width, in_channels].

    :param nb_blocks: 5-D Tensor [batch, numFrames, new height, new width, nb_filter].

    :param bottleneck_size: int`. The number of convolutional filter of the bottleneck convolutional layer.

    :param out_channels: `int`. The number of convolutional filters of the layers surrounding the bottleneck layer.

    :param keep_time_dim: 'boolean'. Tells us if we will keep temporal dimension in this layer.

    :param print_Layer: 'boolean'. Tells us if we will printin this layer.

    :param type: 'char'. Tells us which downsampling strategy we will use.

    :param downsample: `bool`. If True, apply downsampling using 'downsample_strides' for strides.

    :param downsample_strides: int`. The strides to use when downsampling.

    :param activation: `str` (name) or `function` (returning a `Tensor`). Activation applied to this layer (see tflearn.activations). Default: 'linear'.

    :param batch_norm: `bool`. If True, apply batch normalization.

    :param bias: `bool`. If True, a bias is used.

    :param weights_init: `str` (name) or `Tensor`. Weights initialization. (see tflearn.initializations) Default: 'uniform_scaling'.

    :param bias_init: `str` (name) or `tf.Tensor`. Bias initialization. (see tflearn.initializations) Default: 'zeros'.

    :param regularizer: `str` (name) or `Tensor`. Add a regularizer to this layer weights (see tflearn.regularizers). Default: None.

    :param weight_decay: `float`. Regularizer decay parameter. Default: 0.001.

    :param trainable: `bool`. If True, weights will be trainable.

    :param restore: `bool`. If True, this layer weights will be restored when loading a model.

    :param reuse: `bool`. If True and 'scope' is provided, this layer variables will be reused (shared).

    :param scope: `str`. Define this layer scope (optional). A scope can be used to share variables between layers. Note that scope will override name.

    :param name: A name for this layer (optional). Default: 'DeepBottleneck'.

    :return: 5-D Tensor [batch, numFrames, new height, new width, nb_filter].
    """

    resnet = incoming
    in_channels = incoming.get_shape().as_list()[-1]

    with tf.variable_scope(scope, default_name=name, values=[incoming],
                           reuse=reuse) as scope:

        longResidual = 0
        if not downsample:
            longResidual = resnet

        for i in range(nb_blocks):

            identity = resnet

            if print_Layer:
                print resnet.shape

            if not downsample:
                downsample_strides = 1

            if batch_norm:
                resnet = tflearn.batch_normalization(resnet)
            resnet = tflearn.activation(resnet, activation)

            if keep_time_dim:
                resnet = conv_3d(resnet, bottleneck_size, 1,
                                 [1, downsample_strides, downsample_strides], 'valid',
                                 'linear', bias, weights_init,
                                 bias_init, regularizer, weight_decay,
                                 trainable, restore)
            else:
                resnet = conv_3d(resnet, bottleneck_size, 1,
                                 downsample_strides, 'valid',
                                 'linear', bias, weights_init,
                                 bias_init, regularizer, weight_decay,
                                 trainable, restore)

            if batch_norm:
                resnet = tflearn.batch_normalization(resnet)
            resnet = tflearn.activation(resnet, activation)

            resnet = conv_3d(resnet, bottleneck_size, 3, 1, 'same',
                             'linear', bias, weights_init,
                             bias_init, regularizer, weight_decay,
                             trainable, restore)

            resnet = conv_3d(resnet, out_channels, 1, 1, 'valid',
                             activation, bias, weights_init,
                             bias_init, regularizer, weight_decay,
                             trainable, restore)

            # Downsampling
            if downsample_strides > 1:
                # Conv with filter size 1
                if type == 'C':
                    if keep_time_dim:
                        identity = tflearn.conv_3d(identity, out_channels, 1,
                                                   [1, downsample_strides, downsample_strides], 'same', 'linear',
                                                   bias, weights_init, bias_init,
                                                   regularizer, weight_decay, trainable,
                                                   restore)

                    else:
                        identity = tflearn.conv_3d(identity, out_channels, 1,
                                                   downsample_strides, 'same', 'linear',
                                                   bias, weights_init, bias_init,
                                                   regularizer, weight_decay, trainable,
                                                   restore)
                elif type == 'B':

                    if keep_time_dim:
                        identity = tflearn.avg_pool_3d(identity, [1, 1, 2, 2, 1], strides=[1, 2, 2])
                    else:
                        identity = tflearn.avg_pool_3d(identity, downsample_strides,
                                                       downsample_strides)

            # Projection to new dimension
            if type == 'B':
                if in_channels != out_channels:
                    ch = (out_channels - in_channels) // 2
                    identity = tf.pad(identity,
                                      [[0, 0], [0, 0], [0, 0], [0, 0], [ch, ch]])
                    in_channels = out_channels

                resnet = resnet + identity

                resnet = tflearn.activation(resnet, activation)

        if longResidual != 0:
            resChannels = longResidual.get_shape().as_list()[-1]

            if resChannels != out_channels:
                ch = (out_channels - resChannels) // 2
                longResidual = tf.pad(longResidual,
                                  [[0, 0], [0, 0], [0, 0], [0, 0], [ch, ch]])

        resnet = resnet + longResidual

        resnet = tflearn.activation(resnet, activation)

    return resnet



def global_avg_pool3D(incoming, name="GlobalAvgPool3D"):
    """
    3D-Global Average pooling

    :param incoming: 5-D Tensor [batch, numFrames,height, width, in_channels].

    :param name: A name for this layer (optional). Default: 'GlobalAvgPool'.

    :return: 2-D Tensor [batch, pooled dim]
    """

    input_shape = utils.get_incoming_shape(incoming)
    assert len(input_shape) == 5, "Incoming Tensor shape must be 5-D"

    with tf.name_scope(name):
        inference = tf.reduce_mean(incoming, [1, 2, 3])

    # Track output tensor.
    tf.add_to_collection(tf.GraphKeys.LAYER_TENSOR + '/' + name, inference)

    return inference