import glob
import random

import cv2
import numpy as np
from PIL import Image
from sklearn.model_selection import train_test_split

from imageProcessing import zeroCenterOFData, zeroCenterGrayData


class File_Handler(object):
    original_feed_list = []
    original_nonfeed_list = []
    timeStep_feed_list = []
    timeStep_nonfeed_list = []
    numTimesteps = 0
    datasetType = 0

    def __init__(self, feed_folder_path, nonfeed_folder_path, numTimesteps, datasetType):
        self.original_feed_list = []
        self.original_nonfeed_list = []
        self.timeStep_feed_list = []
        self.timeStep_nonfeed_list = []
        self.original_feed_list = self.read_image_list(feed_folder_path)
        self.original_nonfeed_list = self.read_image_list(nonfeed_folder_path)
        self.numTimesteps = numTimesteps
        self.populate_timeStep_lists()
        self.datasetType = datasetType



    def read_image_list(self, folder_path):
        """
        Read image paths from the folder path

        :param folder_path: String
            Path to the folder containing the images
        :return: list of image paths
        """
        image_list = glob.glob(folder_path + "/*.jpg")
        image_list.sort(key=lambda f: int(filter(str.isdigit, f)))

        return image_list

    def shuffle_data(self):
        """
        Shuffles the timestep lists to make them ready for the next batch

        :return: NOTHING
        """
        random.shuffle(self.timeStep_feed_list)
        random.shuffle(self.timeStep_nonfeed_list)

    def populate_timeStep_lists(self):
        """
        Populates the timeStep lists with frames with "numTimesteps" distance appart from the original lists

        --------------------------------------------------------------------------------------------------------------
        Now also creates a deterministic subsampling to enable representative validation and training sets
        --------------------------------------------------------------------------------------------------------------

        :param numTimesteps: int
            Number of timesteps we want between frames
        :return: NOTHING
        """
        tempList = []

        # Populate feeding
        for i in range(len(self.original_feed_list)):
            if ((i%self.numTimesteps == 0) and (i+self.numTimesteps)<len(self.original_feed_list)):
                self.timeStep_feed_list.append(self.original_feed_list[i])

        # Populate nonfeeding
        for j in range(len(self.original_nonfeed_list)):
            if ((j%self.numTimesteps == 0) and (j+self.numTimesteps)<len(self.original_nonfeed_list)):
                tempList.append(self.original_nonfeed_list[j])

        # Since we have a lot more nonfeeding we want to make sure we have the same amount of feeding and nonfeeding
        # Therefore do a subsampling of the nonfeeding timestep list to ensure the same size
        notUsedList, usedList = train_test_split(tempList, test_size = 0.75, random_state = 42)


        # Remove not used list
        del notUsedList[:]

        # Popilate real timeStep_nonfeed_list
        for k in range(len(self.timeStep_feed_list)):
            self.timeStep_nonfeed_list.append(usedList[k])

        # Remove used list as it is not needed anymore
        del usedList[:]



    def get_Batch(self, batch_size, startPos):
        """
        Creates a mini-batch.
        It uses the timeStep lists to create imageCubes of "numTimesteps" consecuteive frames by selecting the "numTimeframes" following frames from the original lists, using path at current index in timeStep list as index in the original list.

        :param batch_size: Size of the mini-batch
        :param startPos: Starting position for the timeStep lists
        :return: mini-batch, labels, new starting position
        """
        feedDatabundle = np.zeros(batch_size/2 * self.numTimesteps * 224 * 224 * 3)
        feedDatabundle = np.reshape(feedDatabundle, (batch_size/2, self.numTimesteps, 224, 224, 3))
        feedImages = np.zeros(self.numTimesteps * 224 * 224 * 3)
        feedImages = np.reshape(feedImages, (self.numTimesteps, 224, 224, 3))

        nonfeedDatabundle = np.zeros(batch_size/2 * self.numTimesteps * 224 * 224 * 3)
        nonfeedDatabundle = np.reshape(nonfeedDatabundle, (batch_size/2, self.numTimesteps, 224, 224, 3))
        nonfeedImages = np.zeros(self.numTimesteps * 224 * 224 * 3)
        nonfeedImages = np.reshape(nonfeedImages, (self.numTimesteps, 224, 224, 3))


        # print len(paths)
        for i in range(batch_size/2):
            feedStartPos = self.original_feed_list.index(self.timeStep_feed_list[startPos])
            nonfeedStartPos = self.original_nonfeed_list.index(self.timeStep_nonfeed_list[startPos])
            startPos +=1
            for j in range(self.numTimesteps):
                # Read in images
                feedImg = np.asarray(cv2.imread(self.original_feed_list[feedStartPos]), dtype=np.float)
                #plt.imshow(feedImg.astype(np.uint8))
                #plt.show()
                nonfeedImg = np.asarray(cv2.imread(self.original_nonfeed_list[nonfeedStartPos]), dtype=np.float)

                # Zero center and normalize the images
                feedImages[j] = zeroCenterOFData(feedImg, self.datasetType)
                nonfeedImages[j] = zeroCenterOFData(nonfeedImg, self.datasetType)


                feedStartPos += 1
                nonfeedStartPos += 1

            # Add images to data bundle
            feedDatabundle[i] = feedImages
            nonfeedDatabundle[i] = nonfeedImages

        feedLabels = np.zeros(shape=(batch_size/2, 1))
        nonfeedLabels = np.zeros(shape=(batch_size / 2, 1))
        for k in range(batch_size/2):
            feedLabels[k] = [1]
            nonfeedLabels[k] = [0]

        imageBatch = np.concatenate((feedDatabundle, nonfeedDatabundle))
        labelsBatch = np.concatenate((feedLabels, nonfeedLabels))

        return imageBatch, labelsBatch, startPos



